# This file is licensed under the terms of the MIT License.
# See the LICENSE file in the root of this repository
# for complete details.

"""Package metadata used by setup.py"""


__all__ = [
    "__title__",
    "__summary__",
    "__url__",
    "__version__",
    "__author__",
    "__email__",
    "__license__",
    "__copyright__",
]

__title__ = "mwsql"
__summary__ = "mwsql is a set of utilities for processing" " MediaWiki SQL dump data."
__url__ = "https://github.com/blancadesal/mwsql"

__version__ = "0.1.5"

__author__ = "Slavina Stefanova"
__email__ = "slavina.stefanova@gmail.com"

__license__ = "MIT License"
__copyright__ = f"Copyright 2021 {__author__}"
